unit MainWindow;

{$MODE OBJFPC}{$LONGSTRINGS ON}{$MODESWITCH TYPEHELPERS}{$MODESWITCH ADVANCEDRECORDS}

interface

uses
  Forms, ExtCtrls, Classes, Controls, Graphics, Types, Math;


type
  TAnchorKind = (anNone, anLeft, anRight, anTop, anBottom, anTopLeft, anTopRight, anBottomLeft, anBottomRight);
  TModeKind = (modNone, modMoving, modResizingHorz, modResizingVert, modResizingDiag);


type
  TRectHelper = record helper for TRect
  private
    function GetMidHorz(): Integer;
    function GetMidVert(): Integer;
  public
    property MidHorz: Integer read GetMidHorz;
    property MidVert: Integer read GetMidVert;
  end;


type
  TSelectionTool = class(TObject)
  private
    FAnchorSize: Integer;
    FMode: TModeKind;
  private
    FArea: TRect;
    FCursorDiff: TPoint;
    FBorderXPtr: PLongInt;
    FBorderYPtr: PLongInt;
  private
    procedure SetAnchorSize(ASize: Integer);
  private
    procedure SwapHorzAreaBorders();
    procedure SwapVertAreaBorders();
  private
    procedure SwapHorzBorderPointer();
    procedure SwapVertBorderPointer();
  private
    function GetActiveHorzAnchors(): Boolean;
    function GetActiveVertAnchors(): Boolean;
  public
    constructor Create(ALeft, ATop, ARight, ABottom: Integer);
  public
    function PointToAnchorRect(AX, AY: Integer): TRect;
    function PointInArea(AX, AY: Integer): Boolean;
  public
    function AnchorFromPoint(AX, AY: Integer): TAnchorKind;
    function CursorForPoint(AX, AY: Integer): TCursor;
  public
    procedure BeginMove(AX, AY: Integer);
    procedure Moving(AX, AY: Integer);
    procedure EndMove();
  public
    procedure BeginResize(AAnchor: TAnchorKind; AX, AY: Integer);
    procedure Resizing(AX, AY: Integer);
    procedure EndResize();
  public
    property AnchorSize: Integer read FAnchorSize write SetAnchorSize;
    property Mode: TModeKind read FMode;
  public
    property Area: TRect read FArea write FArea;
  public
    property ActiveHorzAnchors: Boolean read GetActiveHorzAnchors;
    property ActiveVertAnchors: Boolean read GetActiveVertAnchors;
  end;


type
  TMainForm = class(TForm)
    CPaintBox: TPaintBox;
    procedure FormCreate(ASender: TObject);
    procedure FormDestroy(ASender: TObject);
    procedure FormKeyPress(ASender: TObject; var AKey: Char);
    procedure CPaintBoxPaint(ASender: TObject);
    procedure CPaintBoxMouseMove(ASender: TObject; AShift: TShiftState; AX, AY: Integer);
    procedure CPaintBoxMouseDown(ASender: TObject; AButton: TMouseButton; AShift: TShiftState; AX, AY: Integer);
    procedure CPaintBoxMouseUp(ASender: TObject; AButton: TMouseButton; AShift: TShiftState; AX, AY: Integer);
  private
    FBackgroundImg: TBitmap;
    FSelectionTool: TSelectionTool;
  end;


var
  MainForm: TMainForm;


implementation

{$R *.lfm}


function TRectHelper.GetMidHorz(): Integer;
begin
  Result := Self.Left + Self.Width div 2;
end;


function TRectHelper.GetMidVert(): Integer;
begin
  Result := Self.Top + Self.Height div 2;
end;


constructor TSelectionTool.Create(ALeft, ATop, ARight, ABottom: Integer);
begin
  FAnchorSize := 7;
  FMode := modNone;

  FArea := TRect.Create(ALeft, ATop, ARight, ABottom);
  FCursorDiff := TPoint.Create(0, 0);

  FBorderXPtr := nil;
  FBorderYPtr := nil;
end;


procedure TSelectionTool.SetAnchorSize(ASize: Integer);
begin
  FAnchorSize := Max(3, Min(ASize, 49));
end;


procedure TSelectionTool.SwapHorzAreaBorders();
var
  Location: Integer;
begin
  Location := FArea.Left;
  FArea.Left := FArea.Right;
  FArea.Right := Location;
end;


procedure TSelectionTool.SwapVertAreaBorders();
var
  Location: Integer;
begin
  Location := FArea.Top;
  FArea.Top := FArea.Bottom;
  FArea.Bottom := Location;
end;


procedure TSelectionTool.SwapHorzBorderPointer();
begin
  if FBorderXPtr = @FArea.Left then
    FBorderXPtr := @FArea.Right
  else
    FBorderXPtr := @FArea.Left;
end;


procedure TSelectionTool.SwapVertBorderPointer();
begin
  if FBorderYPtr = @FArea.Top then
    FBorderYPtr := @FArea.Bottom
  else
    FBorderYPtr := @FArea.Top;
end;


function TSelectionTool.GetActiveHorzAnchors(): Boolean;
begin
  Result := Abs(FArea.Height) > FAnchorSize * 2 + 1;
end;


function TSelectionTool.GetActiveVertAnchors(): Boolean;
begin
  Result := Abs(FArea.Width) > FAnchorSize * 2 + 1;
end;


function TSelectionTool.PointToAnchorRect(AX, AY: Integer): TRect;
begin
  Result.Left   := AX - FAnchorSize div 2;
  Result.Top    := AY - FAnchorSize div 2;
  Result.Right  := AX + FAnchorSize div 2 + 1;
  Result.Bottom := AY + FAnchorSize div 2 + 1;
end;


function TSelectionTool.PointInArea(AX, AY: Integer): Boolean;
begin
  Result := FArea.Contains(Point(AX, AY));
end;


function TSelectionTool.AnchorFromPoint(AX, AY: Integer): TAnchorKind;
var
  Location: TPoint;
begin
  Location := TPoint.Create(AX, AY);

  if ActiveHorzAnchors then
  begin
    if PointToAnchorRect(FArea.Left,  FArea.MidVert).Contains(Location) then Exit(anLeft);
    if PointToAnchorRect(FArea.Right, FArea.MidVert).Contains(Location) then Exit(anRight);
  end;

  if ActiveVertAnchors then
  begin
    if PointToAnchorRect(FArea.MidHorz, FArea.Top).Contains(Location)    then Exit(anTop);
    if PointToAnchorRect(FArea.MidHorz, FArea.Bottom).Contains(Location) then Exit(anBottom);
  end;

  if PointToAnchorRect(FArea.Left,  FArea.Top).Contains(Location)    then Exit(anTopLeft);
  if PointToAnchorRect(FArea.Right, FArea.Top).Contains(Location)    then Exit(anTopRight);
  if PointToAnchorRect(FArea.Right, FArea.Bottom).Contains(Location) then Exit(anBottomRight);
  if PointToAnchorRect(FArea.Left,  FArea.Bottom).Contains(Location) then Exit(anBottomLeft);

  Result := anNone;
end;


function TSelectionTool.CursorForPoint(AX, AY: Integer): TCursor;
begin
  if FMode in [modResizingHorz, modResizingVert] then
    Result := Screen.Cursor
  else
  case AnchorFromPoint(AX, AY) of
    anLeft, anRight:          Result := crSizeWE;
    anTop, anBottom:          Result := crSizeNS;
    anTopLeft, anBottomRight: Result := crSizeNWSE;
    anTopRight, anBottomLeft: Result := crSizeNESW;
    anNone:
    case PointInArea(AX, AY) of
      False: Result := crDefault;
      True:  Result := crSizeAll;
    end;
  end;
end;


procedure TSelectionTool.BeginMove(AX, AY: Integer);
begin
  FMode := modMoving;

  FCursorDiff.X := AX - FArea.Left;
  FCursorDiff.Y := AY - FArea.Top;
end;


procedure TSelectionTool.Moving(AX, AY: Integer);
var
  Size: TSize;
begin
  Size := TSize.Create(FArea.Width, FArea.Height);

  FArea.Left := AX - FCursorDiff.X;
  FArea.Top := AY - FCursorDiff.Y;
  FArea.Width := Size.Width;
  FArea.Height := Size.Height;
end;


procedure TSelectionTool.EndMove();
begin
  FMode := modNone;
  FCursorDiff := TPoint.Create(0, 0);
end;


procedure TSelectionTool.BeginResize(AAnchor: TAnchorKind; AX, AY: Integer);
begin
  case AAnchor of
    anLeft, anRight: FMode := modResizingHorz;
    anTop, anBottom: FMode := modResizingVert;
  otherwise
    FMode := modResizingDiag;
  end;

  case FMode of
    modResizingHorz:
    case AAnchor of
      anLeft:  FBorderXPtr := @FArea.Left;
      anRight: FBorderXPtr := @FArea.Right;
    end;
    modResizingVert:
    case AAnchor of
      anTop:    FBorderYPtr := @FArea.Top;
      anBottom: FBorderYPtr := @FArea.Bottom;
    end;
    modResizingDiag:
    begin
      case AAnchor of
        anTopLeft, anBottomLeft:   FBorderXPtr := @FArea.Left;
        anTopRight, anBottomRight: FBorderXPtr := @FArea.Right;
      end;
      case AAnchor of
        anTopLeft, anTopRight:       FBorderYPtr := @FArea.Top;
        anBottomLeft, anBottomRight: FBorderYPtr := @FArea.Bottom;
      end;
    end;
  end;

  if Assigned(FBorderXPtr) then FCursorDiff.X := AX - FBorderXPtr^;
  if Assigned(FBorderYPtr) then FCursorDiff.Y := AY - FBorderYPtr^;
end;


procedure TSelectionTool.Resizing(AX, AY: Integer);
begin
  if Assigned(FBorderXPtr) then FBorderXPtr^ := AX - FCursorDiff.X;
  if Assigned(FBorderYPtr) then FBorderYPtr^ := AY - FCursorDiff.Y;

  if FArea.Left > FArea.Right then
  begin
    SwapHorzAreaBorders();
    SwapHorzBorderPointer();
  end;

  if FArea.Top > FArea.Bottom then
  begin
    SwapVertAreaBorders();
    SwapVertBorderPointer();
  end;
end;


procedure TSelectionTool.EndResize();
begin
  FMode := modNone;

  FBorderXPtr := nil;
  FBorderYPtr := nil;

  FCursorDiff := TPoint.Create(0, 0);
end;


procedure TMainForm.FormCreate(ASender: TObject);
begin
  FBackgroundImg := TBitmap.Create();
  FBackgroundImg.LoadFromFile('back.bmp');

  FSelectionTool := TSelectionTool.Create(100, 100, 200, 200);
end;


procedure TMainForm.FormDestroy(ASender: TObject);
begin
  FBackgroundImg.Free();
  FSelectionTool.Free();
end;


procedure TMainForm.FormKeyPress(ASender: TObject; var AKey: Char);
var
  Location: TPoint;
begin
  case AKey of
    '=', '+':
    begin
      FSelectionTool.AnchorSize := FSelectionTool.AnchorSize + 2;
      CPaintBox.Invalidate();
    end;
    '-', '_':
    begin
      FSelectionTool.AnchorSize := FSelectionTool.AnchorSize - 2;
      CPaintBox.Invalidate();
    end;
  end;

  Location := ScreenToClient(Mouse.CursorPos);
  Screen.Cursor := FSelectionTool.CursorForPoint(Location.X, Location.Y);
end;


procedure TMainForm.CPaintBoxPaint(ASender: TObject);
begin
  with TPaintBox(ASender).Canvas, FSelectionTool do
  begin
    Draw(0, 0, FBackgroundImg);

    Pen.Color   := clBlack;
    Pen.Style   := psDash;
    Brush.Style := bsClear;

    MoveTo(Area.Left,  Area.Top);
    LineTo(Area.Right, Area.Top);
    LineTo(Area.Right, Area.Bottom);
    LineTo(Area.Left,  Area.Bottom);
    LineTo(Area.Left,  Area.Top);

    Pen.Style   := psSolid;
    Brush.Style := bsSolid;
    Brush.Color := clWhite;

    if FSelectionTool.ActiveHorzAnchors then
    begin
      Rectangle(FSelectionTool.PointToAnchorRect(Area.Left,  Area.MidVert));
      Rectangle(FSelectionTool.PointToAnchorRect(Area.Right, Area.MidVert));
    end;

    if FSelectionTool.ActiveVertAnchors then
    begin
      Rectangle(FSelectionTool.PointToAnchorRect(Area.MidHorz, Area.Top));
      Rectangle(FSelectionTool.PointToAnchorRect(Area.MidHorz, Area.Bottom));
    end;

    Rectangle(FSelectionTool.PointToAnchorRect(Area.Left,  Area.Top));
    Rectangle(FSelectionTool.PointToAnchorRect(Area.Right, Area.Top));
    Rectangle(FSelectionTool.PointToAnchorRect(Area.Right, Area.Bottom));
    Rectangle(FSelectionTool.PointToAnchorRect(Area.Left,  Area.Bottom));
  end;
end;


procedure TMainForm.CPaintBoxMouseMove(ASender: TObject; AShift: TShiftState; AX, AY: Integer);
begin
  case FSelectionTool.Mode of
    modMoving:
    begin
      FSelectionTool.Moving(AX, AY);
      CPaintBox.Invalidate();
    end;
    modResizingHorz, modResizingVert, modResizingDiag:
    begin
      FSelectionTool.Resizing(AX, AY);
      CPaintBox.Invalidate();
    end;
  end;

  Screen.Cursor := FSelectionTool.CursorForPoint(AX, AY);
end;


procedure TMainForm.CPaintBoxMouseDown(ASender: TObject; AButton: TMouseButton; AShift: TShiftState; AX, AY: Integer);
var
  Anchor: TAnchorKind;
begin
  if AButton = mbLeft then
  begin
    Anchor := FSelectionTool.AnchorFromPoint(AX, AY);

    if Anchor = anNone then
    begin
      if FSelectionTool.PointInArea(AX, AY) then
      begin
        FSelectionTool.BeginMove(AX, AY);
        CPaintBox.Invalidate();
      end;
    end
    else
    begin
      FSelectionTool.BeginResize(Anchor, AX, AY);
      CPaintBox.Invalidate();
    end;
  end;
end;


procedure TMainForm.CPaintBoxMouseUp(ASender: TObject; AButton: TMouseButton; AShift: TShiftState; AX, AY: Integer);
begin
  if AButton = mbLeft then
  case FSelectionTool.Mode of
    modMoving:
    begin
      FSelectionTool.EndMove();
      CPaintBox.Invalidate();
    end;
    modResizingHorz, modResizingVert, modResizingDiag:
    begin
      FSelectionTool.EndResize();
      CPaintBox.Invalidate();

      Screen.Cursor := FSelectionTool.CursorForPoint(AX, AY);
    end;
  end;
end;


end.

